# -*- coding: utf-8 -*-
""
# Copyright (C) 2014 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# ____________________________________________________________
from tiramisu.i18n import _
from tiramisu.setting import undefined
from tiramisu.error import ConfigError


#____________________________________________________________
#
# Base
class StorageBase(object):
    __slots__ = ('_name', '_requires', '_properties', '_readonly',
                 '_calc_properties', '_informations',
                 '_state_readonly', '_state_requires', '_stated',
                 '_multi', '_validator', '_validator_params',
                 '_default', '_default_multi', '_state_callback', '_callback',
                 '_state_callback_params', '_callback_params', '_multitype',
                 '_consistencies', '_warnings_only', '_master_slaves',
                 '_state_consistencies', '_extra', '_subdyn', '__weakref__',
                 '_state_master_slaves', '_choice_values',
                 '_choice_values_params')

    def __init__(self):
        try:
            self._subdyn
        except AttributeError:
            self._subdyn = None
        try:
            self._consistencies
        except AttributeError:
            self._consistencies = []
        try:
            self._callback
        except AttributeError:
            self._callback = None
        try:
            self._callback_params
        except AttributeError:
            self._callback_params = {}
        try:
            self._validator
        except AttributeError:
            self._validator = None
        try:
            self._validator_params
        except AttributeError:
            self._validator_params = None

    def _add_consistency(self, func, all_cons_opts, params):
        self._consistencies.append((func, all_cons_opts, params))

    def _get_consistencies(self):
        return self._consistencies

    def _get_id(self):
        return id(self)

    def _impl_getsubdyn(self):
        return self._subdyn

    def _impl_setsubdyn(self, subdyn):
        self._subdyn = subdyn

    def commit(self):
        pass


class StorageOptionDescription(StorageBase):
    __slots__ = ('_children', '_cache_paths', '_cache_consistencies',
                 '_group_type', '_is_build_cache', '_state_group_type')

    def __init__(self):
        self._cache_paths = None

    def _add_children(self, child_names, children):
        self._children = (tuple(child_names), tuple(children))

    def impl_already_build_caches(self):
        return self._is_build_cache

    def impl_get_opt_by_path(self, path):
        try:
            return self._cache_paths[0][self._cache_paths[1].index(path)]
        except ValueError:  # pragma: optional cover
            raise AttributeError(_('no option for path {0}').format(path))

    def impl_get_path_by_opt(self, opt):
        try:
            return self._cache_paths[1][self._cache_paths[0].index(opt)]
        except ValueError:  # pragma: optional cover
            raise AttributeError(_('no option {0} found').format(opt))

    def impl_get_group_type(self):  # pragma: optional cover
        return self._group_type

    def impl_build_cache_option(self, _currpath=None, cache_path=None,
                                cache_option=None):
        try:
            self._cache_paths
        except AttributeError:
            self._cache_paths = None
        if _currpath is None and self._cache_paths is not None:  # pragma: optional cover
            # cache already set
            return
        if _currpath is None:
            save = True
            _currpath = []
        else:
            save = False
        if cache_path is None:
            cache_path = []
            cache_option = []
        for option in self._impl_getchildren(dyn=False):
            attr = option._name
            path = str('.'.join(_currpath + [attr]))
            cache_option.append(option)
            cache_path.append(path)
            if option.impl_is_optiondescription():
                _currpath.append(attr)
                option.impl_build_cache_option(_currpath, cache_path,
                                               cache_option)
                _currpath.pop()
        if save:
            self._cache_paths = (tuple(cache_option), tuple(cache_path))

    def impl_get_options_paths(self, bytype, byname, _subpath, only_first, context):
        find_results = []

        def _rebuild_dynpath(path, suffix, dynopt):
            found = False
            spath = path.split('.')
            for length in xrange(1, len(spath)):
                subpath = '.'.join(spath[0:length])
                subopt = self.impl_get_opt_by_path(subpath)
                if dynopt == subopt:
                    found = True
                    break
            if not found:
                raise ConfigError(_('cannot find dynpath'))
            subpath = subpath + suffix
            for slength in xrange(length, len(spath)):
                subpath = subpath + '.' + spath[slength] + suffix
            return subpath

        def _filter_by_name(path, option):
            name = option.impl_getname()
            if option._is_subdyn():
                if byname.startswith(name):
                    found = False
                    for suffix in option._subdyn._impl_get_suffixes(
                            context):
                        if byname == name + suffix:
                            found = True
                            path = _rebuild_dynpath(path, suffix,
                                                    option._subdyn)
                            option = option._impl_to_dyn(
                                name + suffix, path)
                            break
                    if not found:
                        return False
            else:
                if not byname == name:
                    return False
            find_results.append((path, option))
            return True

        def _filter_by_type(path, option):
            if isinstance(option, bytype):
                #if byname is not None, check option byname in _filter_by_name
                #not here
                if byname is None:
                    if option._is_subdyn():
                        name = option.impl_getname()
                        for suffix in option._subdyn._impl_get_suffixes(
                                context):
                            spath = _rebuild_dynpath(path, suffix,
                                                     option._subdyn)
                            find_results.append((spath, option._impl_to_dyn(
                                name + suffix, spath)))
                    else:
                        find_results.append((path, option))
                return True
            return False

        def _filter(path, option):
            if bytype is not None:
                retval = _filter_by_type(path, option)
                if byname is None:
                    return retval
            if byname is not None:
                return _filter_by_name(path, option)

        opts, paths = self._cache_paths
        for index in range(0, len(paths)):
            option = opts[index]
            if option.impl_is_optiondescription():
                continue
            path = paths[index]
            if _subpath is not None and not path.startswith(_subpath + '.'):
                continue
            if bytype == byname is None:
                if option._is_subdyn():
                    name = option.impl_getname()
                    for suffix in option._subdyn._impl_get_suffixes(
                            context):
                        spath = _rebuild_dynpath(path, suffix,
                                                 option._subdyn)
                        find_results.append((spath, option._impl_to_dyn(
                            name + suffix, spath)))
                else:
                    find_results.append((path, option))
            else:
                if _filter(path, option) is False:
                    continue
            if only_first:
                return find_results[0]
        return find_results

    def _impl_st_getchildren(self, context, only_dyn=False):
        for child in self._children[1]:
            if only_dyn is False or child.impl_is_dynoptiondescription():
                yield(child)

    def _getattr(self, name, suffix=undefined, context=undefined, dyn=True):
        error = False
        if suffix is not undefined:
            try:
                if undefined in [suffix, context]:  # pragma: optional cover
                    raise ConfigError(_("suffix and context needed if "
                                        "it's a dyn option"))
                if name.endswith(suffix):
                    oname = name[:-len(suffix)]
                    child = self._children[1][self._children[0].index(oname)]
                    return self._impl_get_dynchild(child, suffix)
                else:
                    error = True
            except ValueError:  # pragma: optional cover
                error = True
        else:
            try:  # pragma: optional cover
                if name == '_readonly':
                    raise AttributeError(_("{0} instance has no attribute "
                                         "'_readonly'").format(
                                             self.__class__.__name__))
                child = self._children[1][self._children[0].index(name)]
                if dyn and child.impl_is_dynoptiondescription():
                    error = True
                else:
                    return child
            except ValueError:
                child = self._impl_search_dynchild(name, context=context)
                if child != []:
                    return child
                error = True
        if error:
            raise AttributeError(_('unknown Option {0} '
                                   'in OptionDescription {1}'
                                   '').format(name, self._name))

    def _get_force_store_value(self):
        #FIXME faire des tests (notamment pas ajouter à un config)
        #FIXME devrait faire un cache !
        opts, paths = self._cache_paths
        for index in range(0, len(paths)):
            opt = opts[index]
            path = paths[index]
            if 'force_store_value' in opt._properties:
                yield (opt, path)
